

#include "clauseset.h"
#include "interpretation.h" 
#include "clause.h"

#include "atom.stat"


#define VERBOSE   1 
   // 0 : means nothing
   // 1 : means new resolved clauses, forward reasoning steps, 
   //     and splitting. 
   // 2 : means everything.


// When there are no false clauses, return s. end( ). 
// Otherwise, return an iterator to a shortest false clause.


clauseset::const_iterator 
findshortestfalseclause( clauseset& s, interpretation& i )
{
   clauseset::const_iterator shortestfalse = s. end( );

   for( clauseset::const_iterator 
           p = s. begin( );
	   p != s. end( );
	   ++ p )
   {
      if( shortestfalse == s. end( ))
      {
         if( p -> cl.isfalse(i))
             shortestfalse = p;
	        // If p -> cl is the first false clause, then it is
		// also the best. 
      }
      else
      {
         if( p -> cl.size( ) < shortestfalse -> cl.size( ) && 
	     p -> cl.isfalse(i))
         {
            shortestfalse = p;
         }
      }
   }
   return shortestfalse;
}



// Finds a clause cl. s.t.
// 1. I does not make the clause cl true. 
// 2. The nr of uninstantiated literals is minimal among such clauses.


clauseset::const_iterator
findnontrueclause( clauseset& s, interpretation& i )
{
   clauseset::const_iterator bestnontrue = s. end( );
   unsigned int bestnr;

   for( clauseset::const_iterator
           p = s. begin( );
	   p != s. end( );
	   ++ p )
   {
      if( ! p -> cl.istrue(i) )
      { 
         unsigned int nr = p -> cl.nrunassigned(i);

         if( bestnontrue == s. end( ))
         {
            bestnontrue = p;
	    bestnr = nr;
	       // If it is the first, then it is also the best.
         }
	 else
         {
            if( nr < bestnr )
            {
               bestnontrue = p;
	       bestnr = nr;
            }
         }
      }
   }

   return bestnontrue; 
}

      


bool sat( clauseset& s, interpretation& i )
{

#if VERBOSE >= 2 
   std::cout  << "Entering SAT with \n" << s << i <<  "\n\n";
#endif

   clauseset::const_iterator c = findnontrueclause( s, i );
   if( c == s. end( ))
   {
      // This means that all clauses are true.
#if VERBOSE >= 1
      std::cout << "all clauses are true\n";
#endif 
      return true;
   }

   unsigned int nr = c -> cl.nrunassigned(i); 
      // If c is not true and has no unassigned literals, then we 
      // can be certain that it is false.

   if( ! nr )
   {
      ASSERT( c -> cl. isfalse(i));
#if VERBOSE >= 1
      std::cout << "found false clause " << ( c -> cl ) << "\n"; 
#endif
      return false; 
   }

   literal lit = c -> cl. findunassigned(i);
   atom at = lit. at;

   if( nr == 1 )
   {
#if VERBOSE >= 1
      std::cout  << "forward reasoning with " << c -> cl << "\n"; 
#endif

      if( lit. positive )
      {
#if VERBOSE >= 1
         std::cout << "assigning " << at << " -> true\n";
#endif 
         i. assign( at, true );
	 bool b = sat(s,i);
	 if(b) return b;
	 i. remove(at);
      }
      else
      {
#if VERBOSE >= 1
         std::cout << "assigning " << at <<  " -> false\n";
#endif
         i. assign( at, false );
	 bool b = sat(s,i);
	 if(b) return b;
	 i. remove(at);
      }
   }
   else
   {
#if VERBOSE >= 1
      std::cout << "splitting on atom " << at << "\n";
      std::cout  << "guessing " << at << " -> false\n"; 
#endif

      i. assign( at, false );
      bool b = sat(s,i);
      if(b) return b;

#if VERBOSE >= 1
      std::cout << "guessing " << at << " -> true\n";
#endif

      i. assign( at, true );
      b = sat(s,i);
      if(b) return b;

      i. remove(at);
   }


   // If we already have a false clause for i, we are done.

   if( findshortestfalseclause(s,i) != s. end( ))
      return false;

   // Otherwise, we have to derive one by resolution:

   i. assign( at, false );
   clauseset::const_iterator f1 = findshortestfalseclause( s, i );
   ASSERT( f1 != s. end( ));

   i. assign( at, true );
   clauseset::const_iterator f2 = findshortestfalseclause( s, i );
   ASSERT( f2 != s. end( ));

   i. remove(at);

   clause res = clause::resolve( at, f1 -> cl, f2 -> cl );
   unsigned int par1 = f1 - s. begin( );
   unsigned int par2 = f2 - s. begin( );

   ASSERT( res. isfalse(i));

#if VERBOSE >= 1
   std::cout << "derived " << s. size( ) << ": " << res; 
   std::cout << " from " << par1  << " and " << par2 << "\n"; 
#endif

   s. insert( res, par1, par2 ); 
   return false;
}





int main(int argc, char **argv)
{


   clauseset set;
   unsigned int cnt = 1; 
   try
   {
      set. read( std::cin, cnt );
   }
   catch( syntaxerror::syntaxerror err )
   {
      std::cout << err << "\n";
      return 0; 
   }
   std::cout << "Input Clauses\n"; 
   std::cout << set; 

   interpretation i;

   if( sat( set, i  ))
   {
      std::cout << i << "\n";
      std::cout << "found satisfying interpretation\n"; 
      return 0; 
   }
   else
   {
      std::cout << set << "\n";
      std::cout << "has no satisfying interpretation\n"; 
      return 0;
   }

}

